export declare enum TokenKind {
    Unknown = "Unknown",
    EndOfFileToken = "EndOfFileToken",
    NameToken = "NameToken",
    ColonToken = "ColonToken",
    MinusToken = "MinusToken",
    DoubleQuoteToken = "DoubleQuoteToken",
    SingleQuoteToken = "SingleQuoteToken",
    OtherToken = "OtherToken",
    MissingToken = "MissingToken",
    SkippedToken = "SkippedToken",
    Expression = "Expression",
    Value = "Value"
}
export declare enum NodeKind {
    FieldNode = "FieldNode",
    FoundToken = "FoundToken",
    InsertedToken = "InsertedToken",
    MissingToken = "MissingToken",
    NegateNode = "NegateNode",
    QueryNode = "QueryNode",
    SkippedTokenNode = "SkippedNode",
    StringLiteralNode = "StringLiteralNode",
    WordNode = "WordNode"
}
/**
 * The entire query, produced from the entire source.
 */
export interface QueryNode {
    readonly kind: NodeKind.QueryNode;
    readonly children: Array<Node>;
}
export interface NegateNode {
    readonly kind: NodeKind.NegateNode;
    readonly minusToken: Token;
    readonly expression: Node | MissingToken;
}
/**
 * A single text word (no spaces).
 */
export interface WordNode {
    readonly kind: NodeKind.WordNode;
    readonly word: Token;
}
export interface StringLiteralNode {
    readonly kind: NodeKind.StringLiteralNode;
    readonly openQuote: Token;
    readonly text: ReadonlyArray<FoundToken> | MissingToken;
    readonly closeQuote: Token;
}
export interface FieldNode {
    readonly kind: NodeKind.FieldNode;
    readonly name: Token;
    readonly colon: Token;
    readonly expression: WordNode | StringLiteralNode | MissingToken;
}
/**
 * An unexpecteed token that was skipped to allow the rest of the source to be
 * parsed.
 */
export interface SkippedTokenNode {
    readonly kind: NodeKind.SkippedTokenNode;
    readonly token: FoundToken;
}
/**
 * An expected token that was actually found in the source (as opposed to a
 * token that was skipped, or a that token that was expected but missing).
 */
export interface FoundToken {
    readonly kind: NodeKind.FoundToken;
    readonly tokenKind: TokenKind;
    readonly fullStart: number;
    readonly start: number;
    readonly length: number;
}
/**
 * An expected token that was not found, but inserted to
 */
export interface MissingToken {
    readonly kind: NodeKind.MissingToken;
    readonly tokenKind: TokenKind;
    readonly fullStart: number;
}
export declare type Token = FoundToken | MissingToken;
export interface InsertedToken {
    readonly kind: NodeKind.InsertedToken;
    readonly text: string;
}
export declare type Node = FieldNode | FoundToken | InsertedToken | MissingToken | MissingToken | NegateNode | QueryNode | SkippedTokenNode | StringLiteralNode | WordNode;
export declare function getChildren(node: Node): ReadonlyArray<Node>;
export declare function getNodeValue(node: WordNode | StringLiteralNode, text: string): string;
export declare function parse(text: string): QueryNode;
export declare function getFullText(node: Node, text: string): string;
export declare function getText(token: Token | Token[], text: string): string;
export declare function toTsQuery(node: Node, text: string): string;
export declare function toPredicate(node: QueryNode, text: string): Predicate;
export declare enum PredicateKind {
    And = "And",
    ContainsText = "ContainsText",
    NoOp = "NoOp",
    Not = "Not",
    ProjectId = "ProjectId",
    ProjectTitleExact = "ProjectTitleExact",
    ProjectTitleFuzzy = "ProjectTitleFuzzy",
    ResultType = "ResultType",
    TagId = "TagId",
    TagTitleExact = "TagTitleExact",
    TagTitleFuzzy = "TagTitleFuzzy"
}
export interface AndPredicate {
    readonly type: PredicateKind.And;
    readonly left: Predicate;
    readonly right: Predicate;
}
export interface ProjectTitleExactPredicate {
    readonly type: PredicateKind.ProjectTitleExact;
    readonly title: string;
}
export interface ProjectTitleFuzzyPredicate {
    readonly type: PredicateKind.ProjectTitleFuzzy;
    readonly title: string;
}
export interface ProjectIdPredicate {
    readonly type: PredicateKind.ProjectId;
    readonly id: string;
}
export interface TagIdPredicate {
    readonly type: PredicateKind.TagId;
    readonly id: string;
}
export interface TagTitleFuzzyPredicate {
    readonly type: PredicateKind.TagTitleFuzzy;
    readonly title: string;
}
export interface TagTitleExactPredicate {
    readonly type: PredicateKind.TagTitleExact;
    readonly title: string;
}
export interface ContainsTextPredicate {
    readonly type: PredicateKind.ContainsText;
    readonly text: string;
}
export interface NoOpPredicate {
    readonly type: PredicateKind.NoOp;
}
export interface NotPredicate {
    readonly type: PredicateKind.Not;
    readonly predicate: Predicate;
}
export interface ResultTypePredicate {
    readonly type: PredicateKind.ResultType;
    readonly resultType: ResultType;
}
export declare type ResultType = "tag" | "insight" | "highlight" | "note";
export declare type Predicate = AndPredicate | NoOpPredicate | ProjectTitleExactPredicate | ProjectTitleFuzzyPredicate | ProjectIdPredicate | TagIdPredicate | TagTitleFuzzyPredicate | TagTitleExactPredicate | ContainsTextPredicate | NotPredicate | ResultTypePredicate;
export declare function refactorTypeTo(node: QueryNode, text: string, type: ResultTypePredicate["resultType"] | null): QueryNode;
export declare function resultTypes(node: QueryNode, text: string): Set<ResultType>;
/**
 * Strip any whitespace from the start of the query by removing it from the
 * first node with whitespace.
 */
export declare function stripLeadingTrivia(node: QueryNode): QueryNode;
