// This compiler is loosely inspired by tolerant-php-parser, specifically
//
// - https://github.com/Microsoft/tolerant-php-parser/blob/master/src/Node.php
// - https://github.com/Microsoft/tolerant-php-parser/blob/master/src/Token.php
export var TokenKind;
(function (TokenKind) {
    TokenKind["Unknown"] = "Unknown";
    TokenKind["EndOfFileToken"] = "EndOfFileToken";
    TokenKind["NameToken"] = "NameToken";
    TokenKind["ColonToken"] = "ColonToken";
    TokenKind["MinusToken"] = "MinusToken";
    TokenKind["DoubleQuoteToken"] = "DoubleQuoteToken";
    TokenKind["SingleQuoteToken"] = "SingleQuoteToken";
    TokenKind["OtherToken"] = "OtherToken";
    TokenKind["MissingToken"] = "MissingToken";
    TokenKind["SkippedToken"] = "SkippedToken";
    // Used for "MissingToken" values
    TokenKind["Expression"] = "Expression";
    TokenKind["Value"] = "Value";
})(TokenKind || (TokenKind = {}));
const textToToken = {
    ":": TokenKind.ColonToken,
    "'": TokenKind.SingleQuoteToken,
    '"': TokenKind.DoubleQuoteToken,
    "-": TokenKind.MinusToken
};
function regexpEscape(str) {
    return str.replace(/[\-\[\]\/\{\}\(\)\*\+\?\.\\\^\$\|]/g, "\\$&");
}
const nonNameCharPattern = new RegExp(`[^${Object.keys(textToToken)
    .map(regexpEscape)
    .join("")}\\s]`);
export var NodeKind;
(function (NodeKind) {
    NodeKind["FieldNode"] = "FieldNode";
    NodeKind["FoundToken"] = "FoundToken";
    NodeKind["InsertedToken"] = "InsertedToken";
    NodeKind["MissingToken"] = "MissingToken";
    NodeKind["NegateNode"] = "NegateNode";
    NodeKind["QueryNode"] = "QueryNode";
    NodeKind["SkippedTokenNode"] = "SkippedNode";
    NodeKind["StringLiteralNode"] = "StringLiteralNode";
    NodeKind["WordNode"] = "WordNode";
})(NodeKind || (NodeKind = {}));
export function getChildren(node) {
    switch (node.kind) {
        case NodeKind.QueryNode:
            return node.children;
        case NodeKind.NegateNode:
            return [node.minusToken, node.expression];
        case NodeKind.WordNode:
            return [node.word];
        case NodeKind.StringLiteralNode:
            return "kind" in node.text
                ? [node.openQuote, node.text, node.closeQuote]
                : [node.openQuote, ...node.text, node.closeQuote];
        case NodeKind.FieldNode:
            return [node.name, node.colon, node.expression];
        case NodeKind.SkippedTokenNode:
            return [node.token];
        case NodeKind.FoundToken:
            return [];
        case NodeKind.MissingToken:
            return [];
        case NodeKind.InsertedToken:
            return [];
    }
}
export function getNodeValue(node, text) {
    switch (node.kind) {
        case NodeKind.WordNode:
            return getText(node.word, text);
        case NodeKind.StringLiteralNode:
            return "kind" in node.text ? "" : node.text.map(token => getFullText(token, text)).join("");
    }
}
class Scanner {
    constructor(text) {
        this.text = text;
        this.pos = 0;
        this.endOfFilePos = this.text.length;
    }
    scanNextToken() {
        const fullStart = this.pos;
        while (true) {
            const start = this.pos;
            if (this.pos >= this.endOfFilePos) {
                return {
                    kind: NodeKind.FoundToken,
                    tokenKind: TokenKind.EndOfFileToken,
                    fullStart: fullStart,
                    start: start,
                    length: this.pos - fullStart
                };
            }
            const char = this.text[this.pos];
            // Skip over whitespace, it contributes to "before trivia"
            if (char.match(/\s/) !== null) {
                this.pos++;
                continue;
            }
            // Check single-char to token mapping
            const kind = textToToken[char];
            if (kind !== undefined) {
                this.pos++;
                return { kind: NodeKind.FoundToken, tokenKind: kind, fullStart: fullStart, start: start, length: this.pos - fullStart };
            }
            let token;
            // Check for a name
            if ((token = this.tryScanName(fullStart, start)) !== null) {
                return token;
            }
            // Capture anything else as an OtherToken
            if ((token = this.tryScanOther(fullStart, start)) !== null) {
                return token;
            }
            throw new Error(`Unexpected character: ${char} charCodeAt(0)=${char.charCodeAt(0)}`);
        }
    }
    tryScanName(fullStart, start) {
        while (this.pos < this.endOfFilePos) {
            const char = this.text[this.pos];
            if (char.match(nonNameCharPattern) !== null) {
                this.pos++;
            }
            else {
                break;
            }
        }
        if (this.pos > start) {
            return {
                kind: NodeKind.FoundToken,
                tokenKind: TokenKind.NameToken,
                fullStart: fullStart,
                start: start,
                length: this.pos - fullStart
            };
        }
        else {
            return null;
        }
    }
    tryScanOther(fullStart, start) {
        if (this.pos < this.endOfFilePos) {
            this.pos++;
            return {
                kind: NodeKind.FoundToken,
                tokenKind: TokenKind.OtherToken,
                fullStart: fullStart,
                start: start,
                length: this.pos - fullStart
            };
        }
        return null;
    }
    lookahead(tokenKind) {
        const originalPos = this.pos;
        const result = this.scanNextToken().tokenKind === tokenKind;
        this.pos = originalPos;
        return result;
    }
}
export function parse(text) {
    const lexer = new Scanner(text);
    let token = lexer.scanNextToken();
    function advanceToken() {
        token = lexer.scanNextToken();
    }
    function eat(tokenKind) {
        if (token.tokenKind === tokenKind) {
            const eaten = token;
            advanceToken();
            return eaten;
        }
        return {
            kind: NodeKind.MissingToken,
            tokenKind: tokenKind,
            fullStart: token.fullStart
        };
    }
    function parseExpression() {
        switch (token.tokenKind) {
            case TokenKind.MinusToken: {
                const minusToken = token;
                advanceToken();
                const node = { kind: NodeKind.NegateNode, minusToken, expression: parsePositiveExpression() };
                return node;
            }
            default: {
                return parsePositiveExpression();
            }
        }
    }
    function parsePositiveExpression() {
        if (token.tokenKind === TokenKind.NameToken) {
            if (lexer.lookahead(TokenKind.ColonToken)) {
                return parseField();
            }
        }
        return parseValue();
    }
    function parseValue() {
        switch (token.tokenKind) {
            case TokenKind.NameToken: {
                const node = { kind: NodeKind.WordNode, word: token };
                advanceToken();
                return node;
            }
            case TokenKind.SingleQuoteToken:
            case TokenKind.DoubleQuoteToken: {
                const quoteTokenKind = token.tokenKind;
                const openQuoteToken = eat(quoteTokenKind);
                const text = [];
                while (token.tokenKind !== quoteTokenKind && token.tokenKind !== TokenKind.EndOfFileToken) {
                    text.push(token);
                    advanceToken();
                }
                const closeQuoteToken = eat(quoteTokenKind);
                const node = {
                    kind: NodeKind.StringLiteralNode,
                    openQuote: openQuoteToken,
                    text: text,
                    closeQuote: closeQuoteToken
                };
                return node;
            }
            default: {
                return {
                    kind: NodeKind.MissingToken,
                    tokenKind: TokenKind.Value,
                    fullStart: token.fullStart
                };
            }
        }
    }
    function parseField() {
        const nameToken = token;
        advanceToken();
        return {
            kind: NodeKind.FieldNode,
            name: nameToken,
            colon: eat(TokenKind.ColonToken),
            expression: parseValue()
        };
    }
    const children = [];
    // Step through each token and attempt to parse.
    while (token.tokenKind !== TokenKind.EndOfFileToken) {
        const startToken = token;
        switch (token.tokenKind) {
            case TokenKind.ColonToken:
            case TokenKind.OtherToken: {
                children.push({ kind: NodeKind.SkippedTokenNode, token });
                advanceToken();
                break;
            }
            default: {
                children.push(parseExpression());
            }
        }
        if (token === startToken) {
            throw new Error(`Failed to parse token ${token.kind}`);
        }
    }
    // Capture the EndOfFileToken.
    children.push(token);
    return { kind: NodeKind.QueryNode, children };
}
export function getFullText(node, text) {
    if (node.kind === NodeKind.FoundToken) {
        return text.substring(node.fullStart, node.fullStart + node.length);
    }
    if (node.kind === NodeKind.InsertedToken) {
        return node.text;
    }
    return getChildren(node)
        .map(nodeOrToken => getFullText(nodeOrToken, text))
        .join("");
}
export function getText(token, text) {
    return Array.isArray(token)
        ? token.map(tok => getText(tok, text)).join("")
        : token.kind === NodeKind.MissingToken
            ? ""
            : text.substring(token.start, token.fullStart + token.length);
}
export function toTsQuery(node, text) {
    if (node.kind === NodeKind.FoundToken) {
        return getText(node, text);
    }
    else if (node.kind === NodeKind.InsertedToken) {
        return "";
    }
    switch (node.kind) {
        case NodeKind.QueryNode: {
            const terms = [];
            for (const child of node.children) {
                const term = toTsQuery(child, text);
                if (term.length > 0) {
                    terms.push(term);
                }
            }
            return terms.join(" & ");
        }
        case NodeKind.WordNode: {
            return getText(node.word, text);
        }
        case NodeKind.NegateNode: {
            const exprText = toTsQuery(node.expression, text);
            return exprText.length > 0 ? `! ${exprText}` : "";
        }
    }
    return "";
}
function isUuid(text) {
    return text.match(/^[a-z0-9]{8}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{12}$/) !== null;
}
export function toPredicate(node, text) {
    const predicates = [];
    function parseTagField(node) {
        const { name, expression } = node;
        if (getText(name, text) === "tag") {
            if (expression.kind === NodeKind.WordNode) {
                const value = getText(expression.word, text);
                return isUuid(value) ? { type: PredicateKind.TagId, id: value } : { type: PredicateKind.TagTitleFuzzy, title: value };
            }
            else if (expression.kind === NodeKind.StringLiteralNode) {
                return { type: PredicateKind.TagTitleExact, title: getNodeValue(expression, text) };
            }
        }
        return;
    }
    function parseTypeField(node) {
        const { name, expression } = node;
        if (getText(name, text) === "type") {
            const value = expression.kind === NodeKind.WordNode || expression.kind === NodeKind.StringLiteralNode
                ? getNodeValue(expression, text)
                : "";
            if (value === "tag" || value === "insight" || value === "highlight" || value === "note") {
                return { type: PredicateKind.ResultType, resultType: value };
            }
        }
        return;
    }
    function parseProjectField(node) {
        const { name, expression } = node;
        if (getText(name, text) === "project") {
            if (expression.kind === NodeKind.WordNode) {
                const value = getText(expression.word, text);
                return isUuid(value)
                    ? { type: PredicateKind.ProjectId, id: value }
                    : { type: PredicateKind.ProjectTitleFuzzy, title: value };
            }
            else if (expression.kind === NodeKind.StringLiteralNode) {
                return { type: PredicateKind.ProjectTitleExact, title: getNodeValue(expression, text) };
            }
        }
        return;
    }
    function compile(child) {
        let value;
        if (child.kind === NodeKind.FieldNode) {
            if ((value = parseTagField(child)) !== undefined ||
                (value = parseProjectField(child)) !== undefined ||
                (value = parseTypeField(child)) !== undefined) {
                return value;
            }
        }
        else if (child.kind === NodeKind.NegateNode) {
            const predicate = compile(child.expression);
            if (predicate !== undefined) {
                return { type: PredicateKind.Not, predicate: predicate };
            }
        }
        else if (child.kind === NodeKind.StringLiteralNode) {
            if (!("kind" in child.text)) {
                const childText = child.text.map(token => getFullText(token, text)).join("");
                if (childText.length > 0) {
                    return {
                        type: PredicateKind.ContainsText,
                        text: childText
                    };
                }
            }
        }
        else if (child.kind === NodeKind.InsertedToken) {
            return;
        }
        else if (child.kind === NodeKind.WordNode) {
            return {
                type: PredicateKind.ContainsText,
                text: getText(child.word, text)
            };
        }
        return;
    }
    for (const child of node.children) {
        const predicate = compile(child);
        if (predicate !== undefined) {
            predicates.push(predicate);
        }
    }
    return predicates.length > 0
        ? predicates.slice(1).reduce((accum, curr) => ({ type: PredicateKind.And, left: accum, right: curr }), predicates[0])
        : { type: PredicateKind.NoOp };
}
export var PredicateKind;
(function (PredicateKind) {
    PredicateKind["And"] = "And";
    PredicateKind["ContainsText"] = "ContainsText";
    PredicateKind["NoOp"] = "NoOp";
    PredicateKind["Not"] = "Not";
    PredicateKind["ProjectId"] = "ProjectId";
    PredicateKind["ProjectTitleExact"] = "ProjectTitleExact";
    PredicateKind["ProjectTitleFuzzy"] = "ProjectTitleFuzzy";
    PredicateKind["ResultType"] = "ResultType";
    PredicateKind["TagId"] = "TagId";
    PredicateKind["TagTitleExact"] = "TagTitleExact";
    PredicateKind["TagTitleFuzzy"] = "TagTitleFuzzy";
})(PredicateKind || (PredicateKind = {}));
export function refactorTypeTo(node, text, type) {
    const children = [];
    for (const child of node.children) {
        let fieldCandidate = child;
        if ("kind" in child && child.kind === NodeKind.NegateNode) {
            fieldCandidate = child.expression;
        }
        if ("kind" in fieldCandidate && fieldCandidate.kind === NodeKind.FieldNode) {
            if (getText(fieldCandidate.name, text) === "type") {
                continue;
            }
        }
        children.push(child);
    }
    if (type !== null) {
        // If the node that will follow the type:___ field starts with
        // non-whitespace, insert some whitespace to separate it from the `type:___`
        // field.
        if (getFullText(children[0], text).match(/^[^\s]/) !== null) {
            children.splice(0, 0, { kind: NodeKind.InsertedToken, text: " " });
        }
        children.splice(0, 0, { kind: NodeKind.InsertedToken, text: `type:${type}` });
        return { kind: NodeKind.QueryNode, children };
    }
    else {
        // Strip any whitespace from the start of the query by removing it from the
        // first node with whitespace.
        return stripLeadingTrivia({ kind: NodeKind.QueryNode, children });
    }
}
const validResultTypes = new Set(["tag", "note", "highlight", "insight"]);
function isValidResultType(type) {
    return validResultTypes.has(type);
}
export function resultTypes(node, text) {
    let requiredTypes;
    let excludedTypes;
    let negate = false;
    const visitor = new class extends Visitor {
        visitNegateNode(node) {
            negate = !negate;
            const result = super.visitNegateNode(node);
            negate = !negate;
            return result;
        }
        visitFieldNode(token) {
            if (getText(token.name, text) === "type" && token.expression.kind !== NodeKind.MissingToken) {
                const type = getNodeValue(token.expression, text);
                if (isValidResultType(type)) {
                    if (negate) {
                        if (excludedTypes === undefined) {
                            excludedTypes = new Set();
                        }
                        excludedTypes.add(type);
                    }
                    else {
                        if (requiredTypes === undefined) {
                            requiredTypes = new Set();
                        }
                        requiredTypes.add(type);
                    }
                }
            }
            return super.visitFieldNode(token);
        }
    }();
    visitor.visitQueryNode(node);
    if (requiredTypes !== undefined && requiredTypes.size > 1) {
        return new Set();
    }
    const result = requiredTypes !== undefined ? requiredTypes : new Set(validResultTypes);
    if (excludedTypes !== undefined) {
        for (const type of excludedTypes) {
            result.delete(type);
        }
    }
    return result;
}
/**
 * Strip any whitespace from the start of the query by removing it from the
 * first node with whitespace.
 */
export function stripLeadingTrivia(node) {
    let foundNonWhitespace = false;
    const stripper = new class extends Visitor {
        visitFoundToken(token) {
            if (!foundNonWhitespace) {
                const leadingTriviaLength = token.start - token.fullStart;
                const newLength = token.length - leadingTriviaLength;
                foundNonWhitespace = newLength > 0;
                if (leadingTriviaLength > 0) {
                    const newLength = token.length - leadingTriviaLength;
                    foundNonWhitespace = newLength > 0;
                    return Object.assign({}, token, { fullStart: token.start, length: newLength });
                }
            }
            return super.visitFoundToken(token);
        }
        visitInsertedToken(token) {
            if (!foundNonWhitespace) {
                const newText = token.text.replace(/^\s+/, "");
                foundNonWhitespace = newText.length > 0;
                return Object.assign({}, token, { text: newText });
            }
            return super.visitInsertedToken(token);
        }
    }();
    return stripper.visitQueryNode(node);
}
class Visitor {
    visitToken(token) {
        if (token.kind === NodeKind.FoundToken) {
            return this.visitFoundToken(token);
        }
        else {
            return this.visitMissingToken(token);
        }
    }
    visitFoundToken(token) {
        return token;
    }
    visitWordNode(node) {
        const newName = this.visitToken(node.word);
        return newName !== node.word ? { kind: node.kind, word: newName } : node;
    }
    visitNegateNode(node) {
        const newMinusToken = this.visitToken(node.minusToken);
        const newExpression = this.visitNode(node.expression);
        return newMinusToken !== node.minusToken || newExpression !== node.expression
            ? { kind: node.kind, minusToken: newMinusToken, expression: newExpression }
            : node;
    }
    visitMissingToken(node) {
        return node;
    }
    visitStringLiteralNode(node) {
        const newOpenQuote = this.visitToken(node.openQuote);
        const newCloseQuote = this.visitToken(node.closeQuote);
        const newTextDraft = "kind" in node.text ? this.visitMissingToken(node.text) : node.text.map(t => this.visitFoundToken(t));
        const textChanged = "kind" in node.text || "kind" in newTextDraft
            ? newTextDraft !== node.text
            : node.text.some((t, i) => t !== newTextDraft[i]);
        const newText = textChanged ? newTextDraft : node.text;
        return newOpenQuote !== node.openQuote || newCloseQuote !== node.closeQuote || newText !== node.text
            ? {
                kind: node.kind,
                openQuote: newOpenQuote,
                text: newText,
                closeQuote: newCloseQuote
            }
            : node;
    }
    visitFieldNode(node) {
        const newName = this.visitToken(node.name);
        const newColon = this.visitToken(node.colon);
        const newExpression = this.visitNode(node.expression);
        return newName !== node.name || newColon !== node.colon || newExpression !== node.expression
            ? {
                kind: node.kind,
                name: newName,
                colon: newColon,
                expression: newExpression
            }
            : node;
    }
    visitQueryNode(node) {
        const newChildrenDraft = node.children.map(child => this.visitNode(child));
        const childrenChanged = node.children.some((child, i) => child !== newChildrenDraft[i]);
        return childrenChanged ? { kind: node.kind, children: newChildrenDraft } : node;
    }
    visitSkippedTokenNode(node) {
        const newToken = this.visitFoundToken(node.token);
        return newToken !== node.token ? { kind: node.kind, token: newToken } : node;
    }
    visitInsertedToken(node) {
        return node;
    }
    visitNode(node) {
        // Hacks due to https://github.com/Microsoft/TypeScript/issues/18215
        switch (node.kind) {
            case NodeKind.FoundToken:
                return this.visitFoundToken(node);
            case NodeKind.MissingToken:
                return this.visitMissingToken(node);
            case NodeKind.InsertedToken:
                return this.visitInsertedToken(node);
            case NodeKind.WordNode:
                return this.visitWordNode(node);
            case NodeKind.NegateNode:
                return this.visitNegateNode(node);
            case NodeKind.StringLiteralNode:
                return this.visitStringLiteralNode(node);
            case NodeKind.FieldNode:
                return this.visitFieldNode(node);
            case NodeKind.QueryNode:
                return this.visitQueryNode(node);
            case NodeKind.SkippedTokenNode:
                return this.visitSkippedTokenNode(node);
        }
    }
}
//# sourceMappingURL=index.js.map