import "jest";
import {
  AndPredicate,
  ContainsTextPredicate,
  getFullText,
  NoOpPredicate,
  NotPredicate,
  parse,
  PredicateKind,
  ProjectIdPredicate,
  ProjectTitleExactPredicate,
  ProjectTitleFuzzyPredicate,
  refactorTypeTo,
  ResultType,
  ResultTypePredicate,
  resultTypes,
  TagTitleExactPredicate,
  TagTitleFuzzyPredicate,
  toPredicate
} from "../index";

describe(parse.name, () => {
  function test(text: string) {
    const query = parse(text);
    expect(query).toMatchSnapshot();
    expect(getFullText(query, text)).toBe(text);
  }

  it("empty", () => test(""));

  it("whitespace", () => test("  "));

  it("-", () => test("-"));

  it("foo", () => test("foo"));

  it(" foo", () => test(" foo"));

  it("foo bar", () => test("foo bar"));

  it("foo  bar", () => test("foo  bar"));

  it("-foo", () => test("-foo"));

  it('"', () => test('"'));

  it("'", () => test("'"));

  it(":", () => test(":"));

  it("foo:", () => test("foo:"));

  it("^", () => test("^"));
});

const b = {
  and: (left: AndPredicate["left"], right: AndPredicate["right"]): AndPredicate => ({
    type: PredicateKind.And,
    left,
    right
  }),
  not: (predicate: NotPredicate["predicate"]): NotPredicate => ({
    type: PredicateKind.Not,
    predicate
  }),
  tagTitleFuzzy: (title: TagTitleFuzzyPredicate["title"]): TagTitleFuzzyPredicate => ({
    type: PredicateKind.TagTitleFuzzy,
    title
  }),
  tagTitleExact: (title: TagTitleExactPredicate["title"]): TagTitleExactPredicate => ({
    type: PredicateKind.TagTitleExact,
    title
  }),
  containsText: (text: ContainsTextPredicate["text"]): ContainsTextPredicate => ({
    type: PredicateKind.ContainsText,
    text
  }),
  projectId: (id: ProjectIdPredicate["id"]): ProjectIdPredicate => ({
    type: PredicateKind.ProjectId,
    id
  }),
  projectTitleFuzzy: (title: ProjectTitleFuzzyPredicate["title"]): ProjectTitleFuzzyPredicate => ({
    type: PredicateKind.ProjectTitleFuzzy,
    title
  }),
  projectTitleExact: (title: ProjectTitleExactPredicate["title"]): ProjectTitleExactPredicate => ({
    type: PredicateKind.ProjectTitleExact,
    title
  }),
  resultType: (resultType: ResultTypePredicate["resultType"]): ResultTypePredicate => ({
    type: PredicateKind.ResultType,
    resultType
  }),
  noOp: (): NoOpPredicate => ({ type: PredicateKind.NoOp })
};

describe(toPredicate.name, () => {
  function test(query: string, expected: {}) {
    expect(toPredicate(parse(query), query)).toMatchObject(expected);
  }

  it("empty", () => test("", {}));

  it("single term", () => test("foo", b.containsText("foo")));

  it("multiple terms", () => test("foo bar", b.and(b.containsText("foo"), b.containsText("bar"))));

  it("quoted multiple terms", () => test('"foo bar"', b.containsText("foo bar")));

  it("single negated term", () => test("-foo", b.not(b.containsText("foo"))));

  it("multiple negated terms", () => test("-foo -bar", b.and(b.not(b.containsText("foo")), b.not(b.containsText("bar")))));

  it("negated with non-negated term", () => test("-foo bar", b.and(b.not(b.containsText("foo")), b.containsText("bar"))));

  it("single -", () => test("-", {}));

  it("single ^", () => test("^", {}));

  it("single !", () => test("!", {}));

  it("tag field", () => test("tag:foo", b.tagTitleFuzzy("foo")));

  it("tag field with quoted value", () => test("tag:'foo bar'", b.tagTitleExact("foo bar")));

  it("tag field with term", () => test("tag:foo bar", b.and(b.tagTitleFuzzy("foo"), b.containsText("bar"))));

  it("negated tag field", () => test("-tag:foo", b.not(b.tagTitleFuzzy("foo"))));

  it("loose negated tag field (space after -)", () => test("- tag:foo", b.not(b.tagTitleFuzzy("foo"))));

  it("multiple negated tag field", () =>
    test("-tag:foo -tag:bar", b.and(b.not(b.tagTitleFuzzy("foo")), b.not(b.tagTitleFuzzy("bar")))));

  it("negated tag field with term", () => test("-tag:foo bar", b.and(b.not(b.tagTitleFuzzy("foo")), b.containsText("bar"))));

  it("negated tag field with negated term", () =>
    test("-tag:foo -bar", b.and(b.not(b.tagTitleFuzzy("foo")), b.not(b.containsText("bar")))));

  it("project field", () => test("project:foo", b.projectTitleFuzzy("foo")));

  it("quoted project field", () => test('project:"foo"', b.projectTitleExact("foo")));

  it("quoted project field with whitespace", () => test('project:"foo bar"', b.projectTitleExact("foo bar")));

  it("multiple project fields", () =>
    test("project:foo project:bar", b.and(b.projectTitleFuzzy("foo"), b.projectTitleFuzzy("bar"))));

  it("type field (tag)", () => test("type:tag", b.resultType("tag")));

  it("type field (insight)", () => test("type:insight", b.resultType("insight")));

  it("type field (highlight)", () => test("type:highlight", b.resultType("highlight")));

  it("type field (note)", () => test("type:note", b.resultType("note")));

  it("unknown type field", () => test("type:x", b.noOp()));
});

describe(refactorTypeTo.name, () => {
  function test(text: string, type: ResultType | null, expected: string) {
    expect(getFullText(refactorTypeTo(parse(text), text, type), text)).toBe(expected);
  }

  it("adds type when non specified", () => test("", "tag", "type:tag"));

  it("preserves unknown fields", () => test("foo:bar", "tag", "type:tag foo:bar"));

  it("preserves known fields", () => test("project:bar", "tag", "type:tag project:bar"));

  it("preserves single terms", () => test("foo", "tag", "type:tag foo"));

  it("preserves multiple terms", () => test("foo bar", "tag", "type:tag foo bar"));

  it("preserves term trivia", () => test("foo   bar ", "tag", "type:tag foo   bar "));

  it("removes existing type fields", () => test("type:insight", "tag", "type:tag"));

  it("removes existing type fields with trivia", () => test("foo   type:insight  bar", "tag", "type:tag foo  bar"));

  it("removes a negated type field", () => test("-type:insight", "tag", "type:tag"));

  it("allows removing a type field", () => test("type:insight", null, ""));

  it("does not introduce extra trailing whitespace", () => test(" foo", "tag", "type:tag foo"));

  it("removes leading trivia from the next token when the field is first", () => test("type:tag  foo", null, "foo"));

  it("preserves leading trivia from the next token when the field is not first", () =>
    test("bar type:tag  foo", null, "bar  foo"));
});

describe(resultTypes.name, () => {
  function test(text: string, expected: Set<string>) {
    expect(resultTypes(parse(text), text)).toMatchObject(expected);
  }

  it("returns all for empty query", () => test("", new Set(["tag", "insight", "highlight", "note"])));

  it("returns tag for `type:tag`", () => test("type:tag", new Set(["tag"])));

  it("returns all but tag for `-type:tag`", () => test("-type:tag", new Set(["insight", "highlight", "note"])));

  it("returns all but tag and note for `-type:tag -type:note`", () =>
    test("-type:tag -type:note", new Set(["insight", "highlight"])));

  it("returns empty for `-type:tag -type:note -type:highlight -type:insight`", () =>
    test("-type:tag -type:note -type:highlight -type:insight", new Set()));

  it("returns empty for `type:tag type:highlight`", () => test("type:tag type:highlight", new Set()));

  it("returns empty for `type:tag -type:tag`", () => test("type:tag -type:tag", new Set()));

  it("returns empty for `type:tag -type:tag type:highlight`", () => test("type:tag -type:tag type:highlight", new Set()));
});
