// This compiler is loosely inspired by tolerant-php-parser, specifically
//
// - https://github.com/Microsoft/tolerant-php-parser/blob/master/src/Node.php
// - https://github.com/Microsoft/tolerant-php-parser/blob/master/src/Token.php

export enum TokenKind {
  Unknown = "Unknown",
  EndOfFileToken = "EndOfFileToken",

  NameToken = "NameToken",
  ColonToken = "ColonToken",
  MinusToken = "MinusToken",
  DoubleQuoteToken = "DoubleQuoteToken",
  SingleQuoteToken = "SingleQuoteToken",
  OtherToken = "OtherToken",
  MissingToken = "MissingToken",
  SkippedToken = "SkippedToken",

  // Used for "MissingToken" values
  Expression = "Expression",
  Value = "Value"
}

const textToToken: { [char: string]: TokenKind | undefined } = {
  ":": TokenKind.ColonToken,
  "'": TokenKind.SingleQuoteToken,
  '"': TokenKind.DoubleQuoteToken,
  "-": TokenKind.MinusToken
};

function regexpEscape(str: string) {
  return str.replace(/[\-\[\]\/\{\}\(\)\*\+\?\.\\\^\$\|]/g, "\\$&");
}

const nonNameCharPattern = new RegExp(
  `[^${Object.keys(textToToken)
    .map(regexpEscape)
    .join("")}\\s]`
);

export enum NodeKind {
  FieldNode = "FieldNode",
  FoundToken = "FoundToken",
  InsertedToken = "InsertedToken",
  MissingToken = "MissingToken",
  NegateNode = "NegateNode",
  QueryNode = "QueryNode",
  SkippedTokenNode = "SkippedNode",
  StringLiteralNode = "StringLiteralNode",
  WordNode = "WordNode"
}

/**
 * The entire query, produced from the entire source.
 */
export interface QueryNode {
  readonly kind: NodeKind.QueryNode;
  readonly children: Array<Node>;
}

export interface NegateNode {
  readonly kind: NodeKind.NegateNode;
  readonly minusToken: Token;
  readonly expression: Node | MissingToken; // Should this be Node | MissingNode?
}

/**
 * A single text word (no spaces).
 */
export interface WordNode {
  readonly kind: NodeKind.WordNode;
  readonly word: Token;
}

export interface StringLiteralNode {
  readonly kind: NodeKind.StringLiteralNode;
  readonly openQuote: Token;
  readonly text: ReadonlyArray<FoundToken> | MissingToken;
  readonly closeQuote: Token;
}

export interface FieldNode {
  readonly kind: NodeKind.FieldNode;
  readonly name: Token;
  readonly colon: Token;
  readonly expression: WordNode | StringLiteralNode | MissingToken;
}

/**
 * An unexpecteed token that was skipped to allow the rest of the source to be
 * parsed.
 */
export interface SkippedTokenNode {
  readonly kind: NodeKind.SkippedTokenNode;
  readonly token: FoundToken;
}

/**
 * An expected token that was actually found in the source (as opposed to a
 * token that was skipped, or a that token that was expected but missing).
 */
export interface FoundToken {
  readonly kind: NodeKind.FoundToken;
  readonly tokenKind: TokenKind;
  readonly fullStart: number;
  readonly start: number;
  readonly length: number;
}

/**
 * An expected token that was not found, but inserted to
 */
export interface MissingToken {
  readonly kind: NodeKind.MissingToken;
  readonly tokenKind: TokenKind;
  readonly fullStart: number;
}

export type Token = FoundToken | MissingToken;

export interface InsertedToken {
  readonly kind: NodeKind.InsertedToken;
  readonly text: string;
}

export type Node =
  | FieldNode
  | FoundToken
  | InsertedToken
  | MissingToken
  | MissingToken
  | NegateNode
  | QueryNode
  | SkippedTokenNode
  | StringLiteralNode
  | WordNode;

export function getChildren(node: Node): ReadonlyArray<Node> {
  switch (node.kind) {
    case NodeKind.QueryNode:
      return node.children;
    case NodeKind.NegateNode:
      return [node.minusToken, node.expression];
    case NodeKind.WordNode:
      return [node.word];
    case NodeKind.StringLiteralNode:
      return "kind" in node.text
        ? [node.openQuote, node.text, node.closeQuote]
        : [node.openQuote, ...node.text, node.closeQuote];
    case NodeKind.FieldNode:
      return [node.name, node.colon, node.expression];
    case NodeKind.SkippedTokenNode:
      return [node.token];
    case NodeKind.FoundToken:
      return [];
    case NodeKind.MissingToken:
      return [];
    case NodeKind.InsertedToken:
      return [];
  }
}

export function getNodeValue(node: WordNode | StringLiteralNode, text: string): string {
  switch (node.kind) {
    case NodeKind.WordNode:
      return getText(node.word, text);
    case NodeKind.StringLiteralNode:
      return "kind" in node.text ? "" : node.text.map(token => getFullText(token, text)).join("");
  }
}

class Scanner {
  private pos = 0;
  private readonly endOfFilePos = this.text.length;

  constructor(private readonly text: string) {}

  public scanNextToken(): FoundToken {
    const fullStart = this.pos;

    while (true) {
      const start = this.pos;

      if (this.pos >= this.endOfFilePos) {
        return {
          kind: NodeKind.FoundToken,
          tokenKind: TokenKind.EndOfFileToken,
          fullStart: fullStart,
          start: start,
          length: this.pos - fullStart
        };
      }

      const char = this.text[this.pos];

      // Skip over whitespace, it contributes to "before trivia"
      if (char.match(/\s/) !== null) {
        this.pos++;
        continue;
      }

      // Check single-char to token mapping
      const kind = textToToken[char];
      if (kind !== undefined) {
        this.pos++;
        return { kind: NodeKind.FoundToken, tokenKind: kind, fullStart: fullStart, start: start, length: this.pos - fullStart };
      }

      let token;

      // Check for a name
      if ((token = this.tryScanName(fullStart, start)) !== null) {
        return token;
      }

      // Capture anything else as an OtherToken
      if ((token = this.tryScanOther(fullStart, start)) !== null) {
        return token;
      }

      throw new Error(`Unexpected character: ${char} charCodeAt(0)=${char.charCodeAt(0)}`);
    }
  }

  private tryScanName(fullStart: number, start: number): FoundToken | null {
    while (this.pos < this.endOfFilePos) {
      const char = this.text[this.pos];
      if (char.match(nonNameCharPattern) !== null) {
        this.pos++;
      } else {
        break;
      }
    }

    if (this.pos > start) {
      return {
        kind: NodeKind.FoundToken,
        tokenKind: TokenKind.NameToken,
        fullStart: fullStart,
        start: start,
        length: this.pos - fullStart
      };
    } else {
      return null;
    }
  }

  private tryScanOther(fullStart: number, start: number): FoundToken | null {
    if (this.pos < this.endOfFilePos) {
      this.pos++;
      return {
        kind: NodeKind.FoundToken,
        tokenKind: TokenKind.OtherToken,
        fullStart: fullStart,
        start: start,
        length: this.pos - fullStart
      };
    }
    return null;
  }

  public lookahead(tokenKind: TokenKind): boolean {
    const originalPos = this.pos;
    const result = this.scanNextToken().tokenKind === tokenKind;
    this.pos = originalPos;
    return result;
  }
}

export function parse(text: string): QueryNode {
  const lexer = new Scanner(text);
  let token = lexer.scanNextToken();

  function advanceToken() {
    token = lexer.scanNextToken();
  }

  function eat(tokenKind: TokenKind): Token {
    if (token.tokenKind === tokenKind) {
      const eaten = token;
      advanceToken();
      return eaten;
    }

    return {
      kind: NodeKind.MissingToken,
      tokenKind: tokenKind,
      fullStart: token.fullStart
    };
  }

  function parseExpression() {
    switch (token.tokenKind) {
      case TokenKind.MinusToken: {
        const minusToken = token;
        advanceToken();
        const node: NegateNode = { kind: NodeKind.NegateNode, minusToken, expression: parsePositiveExpression() };
        return node;
      }

      default: {
        return parsePositiveExpression();
      }
    }
  }

  function parsePositiveExpression(): Node {
    if (token.tokenKind === TokenKind.NameToken) {
      if (lexer.lookahead(TokenKind.ColonToken)) {
        return parseField();
      }
    }

    return parseValue();
  }

  function parseValue(): WordNode | StringLiteralNode | MissingToken {
    switch (token.tokenKind) {
      case TokenKind.NameToken: {
        const node: WordNode = { kind: NodeKind.WordNode, word: token };
        advanceToken();
        return node;
      }

      case TokenKind.SingleQuoteToken:
      case TokenKind.DoubleQuoteToken: {
        const quoteTokenKind = token.tokenKind;
        const openQuoteToken = eat(quoteTokenKind);
        const text = [];
        while (token.tokenKind !== quoteTokenKind && (token as FoundToken).tokenKind !== TokenKind.EndOfFileToken) {
          text.push(token);
          advanceToken();
        }
        const closeQuoteToken = eat(quoteTokenKind);
        const node: StringLiteralNode = {
          kind: NodeKind.StringLiteralNode,
          openQuote: openQuoteToken,
          text: text,
          closeQuote: closeQuoteToken
        };
        return node;
      }

      default: {
        return {
          kind: NodeKind.MissingToken,
          tokenKind: TokenKind.Value,
          fullStart: token.fullStart
        };
      }
    }
  }

  function parseField(): Node {
    const nameToken = token;
    advanceToken();
    return {
      kind: NodeKind.FieldNode,
      name: nameToken,
      colon: eat(TokenKind.ColonToken),
      expression: parseValue()
    };
  }

  const children: QueryNode["children"] = [];
  // Step through each token and attempt to parse.
  while (token.tokenKind !== TokenKind.EndOfFileToken) {
    const startToken = token;
    switch (token.tokenKind) {
      case TokenKind.ColonToken:
      case TokenKind.OtherToken: {
        children.push({ kind: NodeKind.SkippedTokenNode, token });
        advanceToken();
        break;
      }

      default: {
        children.push(parseExpression());
      }
    }

    if (token === startToken) {
      throw new Error(`Failed to parse token ${token.kind}`);
    }
  }

  // Capture the EndOfFileToken.
  children.push(token);

  return { kind: NodeKind.QueryNode, children };
}

export function getFullText(node: Node, text: string): string {
  if (node.kind === NodeKind.FoundToken) {
    return text.substring(node.fullStart, node.fullStart + node.length);
  }

  if (node.kind === NodeKind.InsertedToken) {
    return node.text;
  }

  return getChildren(node)
    .map(nodeOrToken => getFullText(nodeOrToken, text))
    .join("");
}

export function getText(token: Token | Token[], text: string): string {
  return Array.isArray(token)
    ? token.map(tok => getText(tok, text)).join("")
    : token.kind === NodeKind.MissingToken
      ? ""
      : text.substring(token.start, token.fullStart + token.length);
}

export function toTsQuery(node: Node, text: string): string {
  if (node.kind === NodeKind.FoundToken) {
    return getText(node, text);
  } else if (node.kind === NodeKind.InsertedToken) {
    return "";
  }

  switch (node.kind) {
    case NodeKind.QueryNode: {
      const terms = [];
      for (const child of node.children) {
        const term = toTsQuery(child, text);
        if (term.length > 0) {
          terms.push(term);
        }
      }
      return terms.join(" & ");
    }
    case NodeKind.WordNode: {
      return getText(node.word, text);
    }
    case NodeKind.NegateNode: {
      const exprText = toTsQuery(node.expression, text);
      return exprText.length > 0 ? `! ${exprText}` : "";
    }
  }
  return "";
}

function isUuid(text: string) {
  return text.match(/^[a-z0-9]{8}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{12}$/) !== null;
}

export function toPredicate(node: QueryNode, text: string): Predicate {
  const predicates: Predicate[] = [];

  function parseTagField(node: FieldNode): Predicate | undefined {
    const { name, expression } = node;
    if (getText(name, text) === "tag") {
      if (expression.kind === NodeKind.WordNode) {
        const value = getText(expression.word, text);
        return isUuid(value) ? { type: PredicateKind.TagId, id: value } : { type: PredicateKind.TagTitleFuzzy, title: value };
      } else if (expression.kind === NodeKind.StringLiteralNode) {
        return { type: PredicateKind.TagTitleExact, title: getNodeValue(expression, text) };
      }
    }
    return;
  }

  function parseTypeField(node: FieldNode): Predicate | undefined {
    const { name, expression } = node;
    if (getText(name, text) === "type") {
      const value =
        expression.kind === NodeKind.WordNode || expression.kind === NodeKind.StringLiteralNode
          ? getNodeValue(expression, text)
          : "";

      if (value === "tag" || value === "insight" || value === "highlight" || value === "note") {
        return { type: PredicateKind.ResultType, resultType: value };
      }
    }
    return;
  }

  function parseProjectField(node: FieldNode): Predicate | undefined {
    const { name, expression } = node;
    if (getText(name, text) === "project") {
      if (expression.kind === NodeKind.WordNode) {
        const value = getText(expression.word, text);
        return isUuid(value)
          ? { type: PredicateKind.ProjectId, id: value }
          : { type: PredicateKind.ProjectTitleFuzzy, title: value };
      } else if (expression.kind === NodeKind.StringLiteralNode) {
        return { type: PredicateKind.ProjectTitleExact, title: getNodeValue(expression, text) };
      }
    }
    return;
  }

  function compile(child: Node): Predicate | undefined {
    let value;
    if (child.kind === NodeKind.FieldNode) {
      if (
        (value = parseTagField(child)) !== undefined ||
        (value = parseProjectField(child)) !== undefined ||
        (value = parseTypeField(child)) !== undefined
      ) {
        return value;
      }
    } else if (child.kind === NodeKind.NegateNode) {
      const predicate = compile(child.expression);
      if (predicate !== undefined) {
        return { type: PredicateKind.Not, predicate: predicate };
      }
    } else if (child.kind === NodeKind.StringLiteralNode) {
      if (!("kind" in child.text)) {
        const childText = child.text.map(token => getFullText(token, text)).join("");
        if (childText.length > 0) {
          return {
            type: PredicateKind.ContainsText,
            text: childText
          };
        }
      }
    } else if (child.kind === NodeKind.InsertedToken) {
      return;
    } else if (child.kind === NodeKind.WordNode) {
      return {
        type: PredicateKind.ContainsText,
        text: getText(child.word, text)
      };
    }
    return;
  }

  for (const child of node.children) {
    const predicate = compile(child);
    if (predicate !== undefined) {
      predicates.push(predicate);
    }
  }

  return predicates.length > 0
    ? predicates.slice(1).reduce((accum, curr) => ({ type: PredicateKind.And, left: accum, right: curr }), predicates[0])
    : { type: PredicateKind.NoOp };
}

export enum PredicateKind {
  And = "And",
  ContainsText = "ContainsText",
  NoOp = "NoOp",
  Not = "Not",
  ProjectId = "ProjectId",
  ProjectTitleExact = "ProjectTitleExact",
  ProjectTitleFuzzy = "ProjectTitleFuzzy",
  ResultType = "ResultType",
  TagId = "TagId",
  TagTitleExact = "TagTitleExact",
  TagTitleFuzzy = "TagTitleFuzzy"
}

export interface AndPredicate {
  readonly type: PredicateKind.And;
  readonly left: Predicate;
  readonly right: Predicate;
}

export interface ProjectTitleExactPredicate {
  readonly type: PredicateKind.ProjectTitleExact;
  readonly title: string;
}

export interface ProjectTitleFuzzyPredicate {
  readonly type: PredicateKind.ProjectTitleFuzzy;
  readonly title: string;
}

export interface ProjectIdPredicate {
  readonly type: PredicateKind.ProjectId;
  readonly id: string;
}

export interface TagIdPredicate {
  readonly type: PredicateKind.TagId;
  readonly id: string;
}

export interface TagTitleFuzzyPredicate {
  readonly type: PredicateKind.TagTitleFuzzy;
  readonly title: string;
}

export interface TagTitleExactPredicate {
  readonly type: PredicateKind.TagTitleExact;
  readonly title: string;
}

export interface ContainsTextPredicate {
  readonly type: PredicateKind.ContainsText;
  readonly text: string;
}

export interface NoOpPredicate {
  readonly type: PredicateKind.NoOp;
}

export interface NotPredicate {
  readonly type: PredicateKind.Not;
  readonly predicate: Predicate;
}

export interface ResultTypePredicate {
  readonly type: PredicateKind.ResultType;
  readonly resultType: ResultType;
}

export type ResultType = "tag" | "insight" | "highlight" | "note";

export type Predicate =
  | AndPredicate
  | NoOpPredicate
  | ProjectTitleExactPredicate
  | ProjectTitleFuzzyPredicate
  | ProjectIdPredicate
  | TagIdPredicate
  | TagTitleFuzzyPredicate
  | TagTitleExactPredicate
  | ContainsTextPredicate
  | NotPredicate
  | ResultTypePredicate;

export function refactorTypeTo(node: QueryNode, text: string, type: ResultTypePredicate["resultType"] | null): QueryNode {
  const children = [];

  for (const child of node.children) {
    let fieldCandidate = child;
    if ("kind" in child && child.kind === NodeKind.NegateNode) {
      fieldCandidate = child.expression;
    }
    if ("kind" in fieldCandidate && fieldCandidate.kind === NodeKind.FieldNode) {
      if (getText(fieldCandidate.name, text) === "type") {
        continue;
      }
    }
    children.push(child);
  }

  if (type !== null) {
    // If the node that will follow the type:___ field starts with
    // non-whitespace, insert some whitespace to separate it from the `type:___`
    // field.
    if (getFullText(children[0], text).match(/^[^\s]/) !== null) {
      children.splice(0, 0, { kind: NodeKind.InsertedToken, text: " " });
    }
    children.splice(0, 0, { kind: NodeKind.InsertedToken, text: `type:${type}` });
    return { kind: NodeKind.QueryNode, children };
  } else {
    // Strip any whitespace from the start of the query by removing it from the
    // first node with whitespace.
    return stripLeadingTrivia({ kind: NodeKind.QueryNode, children });
  }
}

const validResultTypes = new Set<ResultType>(["tag", "note", "highlight", "insight"]);

function isValidResultType(type: string): type is ResultType {
  return validResultTypes.has(type as ResultType);
}

export function resultTypes(node: QueryNode, text: string): Set<ResultType> {
  let requiredTypes: Set<ResultType> | undefined;
  let excludedTypes: Set<ResultType> | undefined;
  let negate = false;

  const visitor = new class extends Visitor {
    visitNegateNode(node: NegateNode) {
      negate = !negate;
      const result = super.visitNegateNode(node);
      negate = !negate;
      return result;
    }

    visitFieldNode(token: FieldNode) {
      if (getText(token.name, text) === "type" && token.expression.kind !== NodeKind.MissingToken) {
        const type = getNodeValue(token.expression, text);
        if (isValidResultType(type)) {
          if (negate) {
            if (excludedTypes === undefined) {
              excludedTypes = new Set<ResultType>();
            }
            excludedTypes.add(type);
          } else {
            if (requiredTypes === undefined) {
              requiredTypes = new Set<ResultType>();
            }
            requiredTypes.add(type);
          }
        }
      }
      return super.visitFieldNode(token);
    }
  }();

  visitor.visitQueryNode(node);

  if (requiredTypes !== undefined && requiredTypes.size > 1) {
    return new Set();
  }

  const result = requiredTypes !== undefined ? requiredTypes : new Set(validResultTypes);

  if (excludedTypes !== undefined) {
    for (const type of excludedTypes) {
      result.delete(type);
    }
  }

  return result;
}

/**
 * Strip any whitespace from the start of the query by removing it from the
 * first node with whitespace.
 */
export function stripLeadingTrivia(node: QueryNode): QueryNode {
  let foundNonWhitespace = false;

  const stripper = new class extends Visitor {
    visitFoundToken(token: FoundToken) {
      if (!foundNonWhitespace) {
        const leadingTriviaLength = token.start - token.fullStart;
        const newLength = token.length - leadingTriviaLength;

        foundNonWhitespace = newLength > 0;

        if (leadingTriviaLength > 0) {
          const newLength = token.length - leadingTriviaLength;
          foundNonWhitespace = newLength > 0;
          return {
            ...token,
            fullStart: token.start,
            length: newLength
          };
        }
      }
      return super.visitFoundToken(token);
    }

    visitInsertedToken(token: InsertedToken) {
      if (!foundNonWhitespace) {
        const newText = token.text.replace(/^\s+/, "");
        foundNonWhitespace = newText.length > 0;
        return {
          ...token,
          text: newText
        };
      }
      return super.visitInsertedToken(token);
    }
  }();

  return stripper.visitQueryNode(node);
}

class Visitor {
  visitToken(token: Token) {
    if (token.kind === NodeKind.FoundToken) {
      return this.visitFoundToken(token);
    } else {
      return this.visitMissingToken(token);
    }
  }

  visitFoundToken(token: FoundToken): FoundToken {
    return token;
  }

  visitWordNode(node: WordNode): WordNode {
    const newName = this.visitToken(node.word);
    return newName !== node.word ? { kind: node.kind, word: newName } : node;
  }

  visitNegateNode(node: NegateNode): NegateNode {
    const newMinusToken = this.visitToken(node.minusToken);
    const newExpression = this.visitNode(node.expression);
    return newMinusToken !== node.minusToken || newExpression !== node.expression
      ? { kind: node.kind, minusToken: newMinusToken, expression: newExpression }
      : node;
  }

  visitMissingToken(node: MissingToken): MissingToken {
    return node;
  }

  visitStringLiteralNode(node: StringLiteralNode): StringLiteralNode {
    const newOpenQuote = this.visitToken(node.openQuote);
    const newCloseQuote = this.visitToken(node.closeQuote);
    const newTextDraft = "kind" in node.text ? this.visitMissingToken(node.text) : node.text.map(t => this.visitFoundToken(t));

    const textChanged =
      "kind" in node.text || "kind" in newTextDraft
        ? newTextDraft !== node.text
        : node.text.some((t, i) => t !== newTextDraft[i]);

    const newText = textChanged ? newTextDraft : node.text;
    return newOpenQuote !== node.openQuote || newCloseQuote !== node.closeQuote || newText !== node.text
      ? {
          kind: node.kind,
          openQuote: newOpenQuote,
          text: newText,
          closeQuote: newCloseQuote
        }
      : node;
  }

  visitFieldNode(node: FieldNode): FieldNode {
    const newName = this.visitToken(node.name);
    const newColon = this.visitToken(node.colon);
    const newExpression = this.visitNode(node.expression);
    return newName !== node.name || newColon !== node.colon || newExpression !== node.expression
      ? {
          kind: node.kind,
          name: newName,
          colon: newColon,
          expression: newExpression
        }
      : node;
  }

  visitQueryNode(node: QueryNode): QueryNode {
    const newChildrenDraft = node.children.map(child => this.visitNode(child));
    const childrenChanged = node.children.some((child, i) => child !== newChildrenDraft[i]);
    return childrenChanged ? { kind: node.kind, children: newChildrenDraft } : node;
  }

  visitSkippedTokenNode(node: SkippedTokenNode): SkippedTokenNode {
    const newToken = this.visitFoundToken(node.token);
    return newToken !== node.token ? { kind: node.kind, token: newToken } : node;
  }

  visitInsertedToken(node: InsertedToken): InsertedToken {
    return node;
  }

  visitNode<T extends Node>(node: T): T {
    // Hacks due to https://github.com/Microsoft/TypeScript/issues/18215
    switch (node.kind) {
      case NodeKind.FoundToken:
        return this.visitFoundToken(node as FoundToken) as T;
      case NodeKind.MissingToken:
        return this.visitMissingToken(node as MissingToken) as T;
      case NodeKind.InsertedToken:
        return this.visitInsertedToken(node as InsertedToken) as T;
      case NodeKind.WordNode:
        return this.visitWordNode(node as WordNode) as T;
      case NodeKind.NegateNode:
        return this.visitNegateNode(node as NegateNode) as T;
      case NodeKind.StringLiteralNode:
        return this.visitStringLiteralNode(node as StringLiteralNode) as T;
      case NodeKind.FieldNode:
        return this.visitFieldNode(node as FieldNode) as T;
      case NodeKind.QueryNode:
        return this.visitQueryNode(node as QueryNode) as T;
      case NodeKind.SkippedTokenNode:
        return this.visitSkippedTokenNode(node as SkippedTokenNode) as T;
    }
  }
}
